@@ -2,7 +2,9 @@ import logging
 import collections
 import traceback
 import json
+import types
 
+import web
 import gevent
 
 import rpipe.config.protocol
@@ -18,6 +20,8 @@ CONNECTION_CONTEXT_T = collections.namedtuple(
                         'ConnectionContext', 
                         ['participant_address'])
 
+_CT_JSON = 'application/json'
+
 
 class CommonMessageLoop(object):
     def __init__(self, wrapped_socket, event_handler, connection_context):
@@ -105,27 +109,111 @@ class CommonMessageLoop(object):
                      self.__ctx.participant_address, message_obj.verb, 
                      message_obj.noun)
 
-        event_handler_name = message_obj.noun.replace('/', '_')
+        url_parts = message_obj.noun.split('//')
+
+        noun = url_parts[0]
+        if len(url_parts) > 1 and url_parts[1]:
+            parameters = url_parts[1].split('/')
+        else:
+            parameters = []
+
+        handler_parts = [
+            message_obj.verb.lower(),
+            noun.replace('/', '_'),
+        ]
+        
+        event_handler_name = '_'.join(handler_parts)
 
         try:
-            handler = getattr(self.__eh, 'handle_' + event_handler_name)
+            handler = getattr(self.__eh, event_handler_name)
         except AttributeError:
-            _logger.warning("Event is not handled: [%s]", event_handler_name)
+            _logger.warning("Event is not handled: METHOD=[%s]", 
+                            event_handler_name)
 
             self.__send_event_response(
                 message_id, 
                 rpipe.config.protocol.UNHANDLED_EVENT_DEFAULT_RESULT_CODE)
         else:
-            gevent.spawn(
-                self.__process_event,
+            self.__process_event(
                 handler,
                 message_id,
-                message_obj.verb,
-                message_obj.noun,
+                parameters,
+                message_obj.mimetype,
                 message_obj.data)
 
+    def __process_event(self, handler, message_id, parameters, mimetype, data):
+        """Processes event in a new gthread."""
+
+        _logger.debug("Forwarding event to event-handler. MIMETYPE=[%s] "
+                      "PARAMS=%s", mimetype, parameters)
+
+        # We shouldn't even receive data within a GET.
+        if mimetype == _CT_JSON and data:
+            _logger.debug("Decoding JSON data.")
+            data = json.loads(data)
+
+        code = 0
+
+        try:
+            result = handler(self.__ctx, (mimetype, data), *parameters)
+        except rpipe.exceptions.RpHandleException as e:
+            _logger.exception("Handled exception during event: %s", 
+                              e.__class__.__name__)
+
+            result = { 
+                'exception': {
+                    'text': traceback.format_exc(),
+                    'type': 'managed',
+                    'class': e.__class__.__name__,
+                }
+            }
+
+            code = e.code
+        except Exception as e:
+            _logger.exception("Unhandled exception during event: %s", 
+                              e.__class__.__name__)
+
+            result = { 
+                'exception': {
+                    'text': traceback.format_exc(),
+                    'type': 'uncaught',
+                    'class': e.__class__.__name__,
+                }
+            }
+
+            code = rpipe.config.exchange.UNHANDLED_EXCEPTION_CODE
+
+        if issubclass(result.__class__, tuple) is True:
+            (mimetype, code, result_data) = result
+        else:
+            mimetype = None
+            result_data = result
+
+        if mimetype is None:
+            mimetype = _CT_JSON
+
+        _logger.debug("Event result for handler [%s]: [%s] [%s] (%s)", 
+                      handler.func_name, mimetype, 
+                      result_data.__class__.__name__, code)
+
+        if result_data is None:
+            _logger.debug("Result data was None. Coalescing to empty.")
+            result_data = ''
+
+        if issubclass(result_data.__class__, 
+                      (basestring, types.GeneratorType)) is False:
+            if mimetype == _CT_JSON:
+                result_data = json.dumps(result_data)
+            else:
+                raise ValueError("Response to noun [%s] was invalid type and "
+                                 "we're not allowed to encode it to JSON: "
+                                 "[%s]" %
+                                 (noun, result_data.__class__.__name__))
+
+        self.__send_event_response(message_id, code, mimetype, result_data)
+
     def __send_event_response(self, reply_to_message_id, code, 
-                              mimetype='application/json', data={}):
+                              mimetype='text/plain', data=''):
         reply_to_message_id_str = rpipe.protocol.get_string_from_message_id(
                                     reply_to_message_id)
 
@@ -139,29 +227,10 @@ class CommonMessageLoop(object):
         reply_message_obj.version = 1
         reply_message_obj.mimetype = mimetype
         reply_message_obj.code = code
-
-        if mimetype == 'application/json':
-            reply_message_obj.data = json.dumps(data)
-        else:
-            reply_message_obj.data = data
+        reply_message_obj.data = data
 
         rpipe.message_exchange.send(
             self.__ctx.participant_address, 
             reply_message_obj,
             reply_to_message_id=reply_to_message_id,
             expect_response=False)
-
-    def __process_event(self, handler, message_id, verb, noun, data):
-        """Processes event in a new gthread."""
-
-        _logger.debug("Forwarding event to event-handler.")
-
-        try:
-            result_data = handler(self.__ctx, verb, noun, data)
-        except rpipe.exceptions.RpHandleException as e:
-            result_data = traceback.format_exc()
-            code = e.code
-        else:
-            code = 0
-
-        self.__send_event_response(message_id, code, result)
